-- This file uses models and textures taken from TomTom. The 3d arrow model was created by Guillotine (curse.guillotine@gmail.com) and 2d minimap textures by Cladhaire.
local L = LibStub("AceLocale-3.0"):GetLocale("VampArrow")



--[[
      VARIABLES AND UPVALUES ---------------------------------------------------------------------------------------------------
  ]]
  
VampArrow.Arrow = {}

local runAwayArrow
local targetType
local targetPlayer
local targetX, targetY
local hideTime, hideDistance
local ARRIVAL_DISTANCE = 2.0
local showDownArrow = false
local shown = false

local pi, pi2 = math.pi, math.pi * 2
local floor = math.floor
local sin, cos, atan2, sqrt, min = math.sin, math.cos, math.atan2, math.sqrt, math.min
local GetPlayerMapPosition = GetPlayerMapPosition



--[[
      MAP SIZES ----------------------------------------------------------------------------------------------------------------
  ]]
  
VampArrow.MapSizes = {}
	
function VampArrow:RegisterMapSize(zone, ...)
	if not VampArrow.MapSizes[zone] then
		VampArrow.MapSizes[zone] = {}
	end
	local zone = VampArrow.MapSizes[zone]
	for i = 1, select("#", ...), 3 do
		local level, width, height = select(i, ...)
		zone[level] = { width, height }
	end
end



--[[
      FRAME --------------------------------------------------------------------------------------------------------------------
  ]]
  
local frame = CreateFrame("Button", nil, UIParent)
frame:Hide()
frame:SetFrameStrata("FULLSCREEN_DIALOG")
frame:SetWidth(56)
frame:SetHeight(42)
frame:SetMovable(true)
frame:EnableMouse(false)
frame:RegisterForDrag("LeftButton", "RightButton")
frame:SetScript("OnDragStart", function(self)
	self:StartMoving()
end)
frame:SetScript("OnDragStop", function(self)
	self:StopMovingOrSizing()
	local point, _, _, x, y = self:GetPoint(1)
	VampArrow.db.profile.arrowPoint = point
	VampArrow.db.profile.arrowPosX = x
	VampArrow.db.profile.arrowPosY = y
end)
local arrow = frame:CreateTexture(nil, "OVERLAY")
arrow:SetTexture("Interface\\AddOns\\VampArrow\\Resources\\Arrow.blp")
arrow:SetAllPoints(frame)

local titleframe = CreateFrame("Frame", nil, frame)
frame.title = titleframe:CreateFontString("OVERLAY", nil, "GameFontHighlightSmall")
frame.status = titleframe:CreateFontString("OVERLAY", nil, "GameFontNormalSmall")
frame.title:SetPoint("TOP", frame, "BOTTOM", 0, 0)
frame.status:SetPoint("TOP", frame.title, "BOTTOM", 0, 0)



--[[
      UTILITIES ----------------------------------------------------------------------------------------------------------------
  ]]
  

local SetMapToCurrentZone -- throttled SetMapToCurrentZone function to prevent lag issues with unsupported WorldMap addons
do
	local lastMapUpdate = 0
	function SetMapToCurrentZone(...)
		if GetTime() - lastMapUpdate > 1 then
			lastMapUpdate = GetTime()
			return _G.SetMapToCurrentZone(...)
		end
	end
end


local calculateDistance
do
	local mapSizes = VampArrow.MapSizes
	function calculateDistance(x1, y1, x2, y2)
		local mapName = GetMapInfo()
		local levels = mapSizes[mapName]
		if not levels then
			return
		end
		local dims = levels[GetCurrentMapDungeonLevel()]
		if not dims and levels and GetCurrentMapDungeonLevel() == 0 then -- we are in a known zone but the dungeon level seems to be wrong
			SetMapToCurrentZone() -- fixes the dungeon level (if it was wrong for some reason)
			dims = levels[GetCurrentMapDungeonLevel()] -- try again
		end
		if not dims then -- we are in an unknown dungeon :(
			return
		end
		local dX = (x1 - x2) * dims[1]
		local dY = (y1 - y2) * dims[2]
		return sqrt(dX * dX + dY * dY)
	end
end


-- GetPlayerFacing seems to return values between -pi and pi instead of 0 - 2pi sometimes since 3.3.3
local GetPlayerFacing = function(...)
	local result = GetPlayerFacing(...)
	if result < 0 then
		result = result + pi2
	end
	return result
end


local function ColorGradient(perc, ...)
	local num = select("#", ...)
	local hexes = type(select(1, ...)) == "string"

	if perc == 1 then
		return select(num-2, ...), select(num-1, ...), select(num, ...)
	end

	num = num / 3

	local segment, relperc = math.modf(perc*(num-1))
	local r1, g1, b1, r2, g2, b2
	r1, g1, b1 = select((segment*3)+1, ...), select((segment*3)+2, ...), select((segment*3)+3, ...)
	r2, g2, b2 = select((segment*3)+4, ...), select((segment*3)+5, ...), select((segment*3)+6, ...)

	if not r2 or not g2 or not b2 then
		return r1, g1, b1
	else
		return r1 + (r2-r1)*relperc,
		g1 + (g2-g1)*relperc,
		b1 + (b2-b1)*relperc
	end
end


local updateArrow
do
	local currentCell
	local count
	
	function updateArrow(direction, distance)
		if distance and distance <= ARRIVAL_DISTANCE and targetType ~= "rotate" then
		
			if not showDownArrow then
				arrow:SetHeight(70)
				arrow:SetWidth(53)
				arrow:SetTexture("Interface\\AddOns\\VampArrow\\Resources\\Arrow-UP.blp")
				arrow:SetVertexColor(0, 1, 0, 1)
				showDownArrow = true
			end

			if not count then count = 0 end
			count = count + 1
			if count >= 55 then
				count = 0
			end

			cell = count
			local column = cell % 9
			local row = floor(cell / 9)

			local xstart = (column * 53) / 512
			local ystart = (row * 70) / 512
			local xend = ((column + 1) * 53) / 512
			local yend = ((row + 1) * 70) / 512
			arrow:SetTexCoord(xstart, xend, ystart, yend)
			
		else
		
			if showDownArrow then
				arrow:SetHeight(56)
				arrow:SetWidth(42)
				arrow:SetTexture("Interface\\AddOns\\VampArrow\\Resources\\Arrow.blp")
				showDownArrow = false
			end

			local cell = floor(direction / pi2 * 108 + 0.5) % 108
			if cell ~= currentCell then
				currentCell = cell
				local column = cell % 9
				local row = floor(cell / 9)
				local xStart = (column * 56) / 512
				local yStart = (row * 42) / 512
				local xEnd = ((column + 1) * 56) / 512
				local yEnd = ((row + 1) * 42) / 512
				arrow:SetTexCoord(xStart, xEnd, yStart, yEnd)
			end
			
			local perc = math.abs((math.pi - math.abs(direction)) / math.pi)
			if runAwayArrow then
				local br, bg, bb = 1, 0, 0 -- bad
				local mr, mg, mb = 1, 1, 0 -- middle
				local gr, gg, gb = 0, 1, 0 -- good
				local r, g, b = ColorGradient(perc, br, bg, bb, mr, mg, mb, gr, gg, gb)		
				arrow:SetVertexColor(r, g, b)
			else
				local gr, gg, gb = 0, 1, 0 -- good
				local mr, mg, mb = 1, 1, 0 -- middle
				local br, bg, bb = 1, 0, 0 -- bad
				local r, g, b = ColorGradient(perc, br, bg, bb, mr, mg, mb, gr, gg, gb)		
				arrow:SetVertexColor(r, g, b)
			end
				
			if distance then
				if runAwayArrow then
					if hideDistance and distance >= hideDistance then
						frame:Hide()
					end
				else
					if hideDistance and distance <= hideDistance then
						frame:Hide()
					end
				end
			end
			
		end
	end
	
end



--[[
      ONUPDATE HANDLER ---------------------------------------------------------------------------------------------------------
  ]]
  
do
	local rotateState = 0
	frame:SetScript("OnUpdate", function(self, elapsed)
		if not shown then return end
		
		if WorldMapFrame:IsShown() then -- it doesn't work while the world map frame is shown
			arrow:Hide()
			return
		end
		if hideTime and GetTime() > hideTime then
			frame:Hide()
		end
		arrow:Show()
		local x, y = GetPlayerMapPosition("player")
		if x == 0 and y == 0 then
			SetMapToCurrentZone()
			x, y = GetPlayerMapPosition("player")
			if x == 0 and y == 0 then
				self:Hide() -- hide the arrow if you enter a zone without a map
				return
			end
		end
		if targetType == "player" then
			targetX, targetY = GetPlayerMapPosition(targetPlayer)
			if targetX == 0 and targetY == 0 then
				self:Hide() -- hide the player if the target doesn't exist. TODO: just hide the texture and add a timeout
			end
			local targetText = targetPlayer
			if VampArrow.Arrow.mark and VampArrow.Arrow.mark >= 1 and VampArrow.Arrow.mark <= 8 then
				targetText = targetText.." "..ICON_LIST[VampArrow.Arrow.mark].."0|t"..ICON_LIST[VampArrow.Arrow.mark].."0|t"..ICON_LIST[VampArrow.Arrow.mark].."0|t"
			end
			frame.title:SetText(targetText)
		elseif targetType == "rotate" then
			rotateState = rotateState + elapsed
			targetX = x + cos(rotateState)
			targetY = y + sin(rotateState)
		end
		if not targetX or not targetY then
			return
		end
		local angle = atan2(x - targetX, targetY - y)
		if angle <= 0 then -- -pi < angle < pi but we need/want a value between 0 and 2 pi
			if runAwayArrow then
				angle = -angle -- 0 < angle < pi
			else
				angle = pi - angle -- pi < angle < 2pi
			end
		else
			if runAwayArrow then
				angle = pi2 - angle -- pi < angle < 2pi
			else
				angle = pi - angle  -- 0 < angle < pi
			end
		end
		local distance = calculateDistance(x, y, targetX, targetY)
		if targetType == "player" and distance then
			frame.status:SetText(format("%.2f "..L["yards"], distance))
		end
		updateArrow(angle - GetPlayerFacing(), distance)
	end)
end



--[[
      API ----------------------------------------------------------------------------------------------------------------------
  ]]
  
local function show(runAway, x, y, distance, time)
	shown = true
	local player
	if type(x) == "string" then
		player, hideDistance, hideTime = x, y, hideDistance
	end
	frame:Show()
	runAwayArrow = runAway
	--hideDistance = distance or runAway and 100 or 3
	if time then
		hideTime = time + GetTime()
	else
		hideTime = nil
	end
	if player then
		targetType = "player"
		targetPlayer = player
	else
		targetType = "fixed"
		targetX, targetY = x, y
	end
	if not VampArrow.db.profile.lockArrow then
		frame:EnableMouse(true)
	end
end

function VampArrow.Arrow:ShowRunTo(...)
	return show(false, ...)
end

function VampArrow.Arrow:ShowRunAway(...)
	return show(true, ...)
end

function VampArrow.Arrow:Hide(autoHide)
	shown = false
	frame:Hide()
end

function VampArrow.Arrow:EndMove()
	frame:EnableMouse(false)
	VampArrow.Arrow:Hide()
end

function VampArrow.Arrow:Move()
	shown = true
	targetType = "rotate"
	runAwayArrow = false
	frame.title:SetText(L["Move the arrow"])
	frame.status:SetText("")
	frame:EnableMouse(true)
	frame:Show()
end

function VampArrow.Arrow:LoadPosition()
	frame:SetPoint(VampArrow.db.profile.arrowPoint or "TOP", VampArrow.db.profile.arrowPosX or 0, VampArrow.db.profile.arrowPosY or -150)
end
